from typing import ClassVar

from obi_one.core.block import Block
from obi_one.core.form import Form
from obi_one.core.single import SingleCoordinateMixin
from obi_one.scientific.circuit.circuit import Circuit
from obi_one.scientific.circuit.neuron_sets import NeuronSet
from obi_one.scientific.unions.unions_neuron_sets import NeuronSetUnion


class CircuitExtractions(Form):
    """ """

    single_coord_class_name: ClassVar[str] = "CircuitExtraction"
    name: ClassVar[str] = "Circuit Extraction"
    description: ClassVar[str] = (
        "Extracts a sub-circuit of a SONATA circuit as defined by a node set. The output circuit will contain all morphologies, hoc files, and mod files that are required to simulate the extracted circuit."
    )
    
    class Initialize(Block):
        circuit: Circuit | list[Circuit]
        run_validation: bool = False

    initialize: Initialize
    neuron_set: NeuronSetUnion

    def save_collection(self, circuit_entities):
        pass
        """
        Should save the collaction object here
        """


import json
import os
import shutil
import traceback

import bluepysnap as snap
import bluepysnap.circuit_validation
import h5py
import tqdm
from brainbuilder.utils.sonata import split_population


class CircuitExtraction(CircuitExtractions, SingleCoordinateMixin):
    """Extracts a sub-circuit of a SONATA circuit as defined by a node set. The output circuit will contain
    all morphologies, hoc files, and mod files that are required to simulate the extracted circuit.
    """

    @staticmethod
    def _filter_ext(file_list, ext):
        return list(filter(lambda f: os.path.splitext(f)[1].lower() == f".{ext}", file_list))

    def run(self) -> str:
        try:
            # Add neuron set to SONATA circuit object
            # (will raise an error in case already existing)
            nset_name = self.neuron_set.__class__.__name__
            nset_def = self.neuron_set.get_node_set_definition(
                self.initialize.circuit, self.initialize.circuit.default_population_name
            )
            sonata_circuit = self.initialize.circuit.sonata_circuit
            NeuronSet.add_node_set_to_circuit(
                sonata_circuit, {nset_name: nset_def}, overwrite_if_exists=False
            )

            # Create subcircuit using "brainbuilder"
            print(f"Extracting subcircuit from '{self.initialize.circuit.name}'")
            split_population.split_subcircuit(
                self.coordinate_output_root,
                nset_name,
                sonata_circuit,
                True,
                False,
            )

            # Custom edit of the circuit config so that all paths are relative to the new base directory
            # (in case there were absolute paths in the original config)
            def rebase_config(config_dict, old_base, new_base):
                for key, value in config_dict.items():
                    if isinstance(value, str):
                        if value == old_base:
                            config_dict[key] = ""
                        else:
                            config_dict[key] = value.replace(old_base, new_base)
                    elif isinstance(value, dict):
                        rebase_config(value, old_base, new_base)
                    elif isinstance(value, list):
                        for _v in value:
                            rebase_config(_v, old_base, new_base)

            old_base = os.path.split(self.initialize.circuit.path)[0]

            # Quick fix to deal with symbolic links in base circuit (not usually required)
            # alt_base = old_base  # Alternative old base
            # for _sfix in ["-ER", "-DD", "-BIP", "-OFF", "-POS"]:
            #     alt_base = alt_base.removesuffix(_sfix)

            new_base = "$BASE_DIR"
            new_circuit_path = os.path.join(self.coordinate_output_root, "circuit_config.json")
            # shutil.copyfile(new_circuit_path, os.path.splitext(new_circuit_path)[0] + ".BAK")  # Create backup before modifying

            with open(new_circuit_path) as config_file:
                config_dict = json.load(config_file)
            rebase_config(config_dict, old_base, new_base)
            # if alt_base != old_base:  # Quick fix to deal with symbolic links in base circuit
            # rebase_config(config_dict, alt_base, new_base)
            with open(new_circuit_path, "w") as config_file:
                json.dump(config_dict, config_file, indent=4)

            # Copy subcircuit morphologies and e-models (separately per node population)
            original_circuit = self.initialize.circuit.sonata_circuit
            new_circuit = snap.Circuit(new_circuit_path)
            for pop_name, pop in new_circuit.nodes.items():
                if pop.config["type"] == "biophysical":
                    # Copying morphologies of any (supported) format
                    if "morphology" in pop.property_names:
                        print(f"Copying morphologies for population '{pop_name}' ({pop.size})")
                        morphology_list = pop.get(properties="morphology").unique()

                        src_morph_dirs = {}
                        dest_morph_dirs = {}
                        for _morph_ext in ["swc", "asc", "h5"]:
                            try:
                                morph_folder = original_circuit.nodes[
                                    pop_name
                                ].morph._get_morphology_base(_morph_ext)
                                # FIXME: Should not use private function!! But required to get path even if h5 container.

                                assert os.path.exists(morph_folder), (
                                    f"ERROR: {_morph_ext} morphology folder/container does not exist!"
                                )

                                if os.path.isdir(morph_folder):
                                    assert (
                                        len(self._filter_ext(os.listdir(morph_folder), _morph_ext))
                                        > 0
                                    ), (
                                        f"ERROR: {_morph_ext} morphology folder does not contain morphologies!"
                                    )

                                dest_morph_dirs[_morph_ext] = pop.morph._get_morphology_base(
                                    _morph_ext
                                )  # FIXME: Should not use private function!!
                                src_morph_dirs[_morph_ext] = morph_folder
                            except:
                                morph_folder = None

                        assert len(src_morph_dirs) > 0, (
                            "ERROR: No morphologies of any supported format found!"
                        )
                        for _morph_ext in src_morph_dirs:
                            if _morph_ext == "h5" and os.path.isfile(src_morph_dirs[_morph_ext]):
                                # FIXME: If there is only one neuron extracted, consider removing the container!!
                                # Copy containerized morphologies into new container
                                os.makedirs(
                                    os.path.split(dest_morph_dirs[_morph_ext])[0], exist_ok=True
                                )
                                src_container = src_morph_dirs[_morph_ext]
                                dest_container = dest_morph_dirs[_morph_ext]
                                with h5py.File(src_container) as f_src:
                                    with h5py.File(dest_container, "a") as f_dest:
                                        skip_counter = 0
                                        for morphology_name in tqdm.tqdm(
                                            morphology_list,
                                            desc=f"Copying containerized .{_morph_ext} morphologies",
                                        ):
                                            if morphology_name in f_dest:
                                                skip_counter += 1
                                            else:
                                                f_src.copy(
                                                    f_src[morphology_name],
                                                    f_dest,
                                                    name=morphology_name,
                                                )
                                print(
                                    f"Copied {len(morphology_list) - skip_counter} morphologies into container ({skip_counter} already existed)"
                                )
                            else:
                                # Copy morphology files
                                os.makedirs(dest_morph_dirs[_morph_ext], exist_ok=True)
                                for morphology_name in tqdm.tqdm(
                                    morphology_list, desc=f"Copying .{_morph_ext} morphologies"
                                ):
                                    src_file = os.path.join(
                                        src_morph_dirs[_morph_ext],
                                        f"{morphology_name}.{_morph_ext}",
                                    )
                                    dest_file = os.path.join(
                                        dest_morph_dirs[_morph_ext],
                                        f"{morphology_name}.{_morph_ext}",
                                    )
                                    assert os.path.exists(src_file), (
                                        f"ERROR: Morphology '{src_file}' missing!"
                                    )
                                    if not os.path.exists(dest_file):
                                        # Copy only, if not yet existing (could happen for shared morphologies among populations)
                                        shutil.copyfile(src_file, dest_file)

                    # Copy .hoc file directory (Even if defined globally, shows up under pop.config)
                    if "biophysical_neuron_models_dir" in pop.config:
                        hoc_file_list = [
                            _hoc.split(":")[-1] + ".hoc"
                            for _hoc in pop.get(properties="model_template").unique()
                        ]
                        print(
                            f"Copying {len(hoc_file_list)} biophysical neuron models (.hoc) for population '{pop_name}' ({pop.size})"
                        )

                        source_dir = original_circuit.nodes[pop_name].config[
                            "biophysical_neuron_models_dir"
                        ]
                        dest_dir = pop.config["biophysical_neuron_models_dir"]
                        os.makedirs(dest_dir, exist_ok=True)

                        for _hoc_file in hoc_file_list:
                            src_file = os.path.join(source_dir, _hoc_file)
                            dest_file = os.path.join(dest_dir, _hoc_file)
                            assert os.path.exists(src_file), (
                                f"ERROR: HOC file '{src_file}' missing!"
                            )
                            if not os.path.exists(dest_file):
                                # Copy only, if not yet existing (could happen for shared hoc files among populations)
                                shutil.copyfile(src_file, dest_file)

            # Copy .mod files, if any
            mod_folder = "mod"
            source_dir = os.path.join(os.path.split(self.initialize.circuit.path)[0], mod_folder)
            if os.path.exists(source_dir):
                print("Copying mod files")
                dest_dir = os.path.join(self.coordinate_output_root, mod_folder)
                shutil.copytree(source_dir, dest_dir)

            # Run circuit validation
            if self.initialize.run_validation:
                errors = snap.circuit_validation.validate(new_circuit_path, skip_slow=True)
                assert len(errors) == 0, f"Circuit validation error(s) found: {errors}"
                print("No validation errors found!")

            print("Extraction DONE")

        except Exception as e:
            if str(e) == "Unable to synchronously create group (name already exists)":
                print(
                    "Error:",
                    f"Subcircuit {self.neuron_set} already exists. Subcircuit must be deleted before running the extraction.",
                )
            else:
                traceback.print_exception(e)
            return

    def save(self):
        pass
        """
        Currently should return a created entity
        """



import copy
import json
import logging
from collections import OrderedDict
from importlib.metadata import version
from itertools import product
from pathlib import Path
import entitysdk

from pydantic import PrivateAttr, ValidationError

from obi_one.core.base import OBIBaseModel
from obi_one.core.block import Block
from obi_one.core.param import MultiValueScanParam, SingleValueScanParam
from obi_one.core.single import SingleCoordinateMixin, SingleCoordinateScanParams
from obi_one.scientific.unions.unions_form import FormUnion

L = logging.getLogger(__name__)


class Scan(OBIBaseModel):
    """Takes a Form & output_root as input.

    - Creates multi-dimensional parameter scans through calls to generate and run
    - Includes several intermediate functions for computing multi-dimensional parameter scans:
        i.e. multiple_value_parameters, coordinate_parameters, coordinate_instances
    """

    form: FormUnion
    output_root: Path = Path()
    coordinate_directory_option: str = "NAME_EQUALS_VALUE"
    _multiple_value_parameters: list = None
    _coordinate_parameters: list = PrivateAttr(default=[])
    _coordinate_instances: list = PrivateAttr(default=[])

    @property
    def output_root_absolute(self) -> Path:
        """Returns the absolute path of the output_root."""
        print(self.output_root.resolve())
        return self.output_root.resolve()

    def multiple_value_parameters(self, *, display: bool = False) -> list[MultiValueScanParam]:
        """Iterates through Blocks of self.form to find "multi value parameters".

            (i.e. parameters with list values of length greater than 1)
        - Returns a list of MultiValueScanParam objects
        """
        self._multiple_value_parameters = []

        # Iterate through all attributes of the Form
        for attr_name, attr_value in self.form.__dict__.items():
            # Check if the attribute is a dictionary of Block instances
            if isinstance(attr_value, dict) and all(
                isinstance(dict_val, Block) for dict_key, dict_val in attr_value.items()
            ):
                category_name = attr_name
                category_blocks_dict = attr_value

                # If so iterate through the dictionary's Block instances
                for block_key, block in category_blocks_dict.items():
                    # Call the multiple_value_parameters method of the Block instance
                    block_multi_value_parameters = block.multiple_value_parameters(
                        category_name=category_name, block_key=block_key
                    )
                    if len(block_multi_value_parameters):
                        self._multiple_value_parameters.extend(block_multi_value_parameters)

            # Else if the attribute is a Block instance, call the _multiple_value_parameters method
            # of the Block instance
            if isinstance(attr_value, Block):
                block_name = attr_name
                block = attr_value
                block_multi_value_parameters = block.multiple_value_parameters(
                    category_name=block_name
                )
                if len(block_multi_value_parameters):
                    self._multiple_value_parameters.extend(block_multi_value_parameters)

        # Optionally display the multiple_value_parameters
        if display:
            L.info("\nMULTIPLE VALUE PARAMETERS")
            if len(self._multiple_value_parameters) == 0:
                L.info("No multiple value parameters found.")
            else:
                for multi_value in self._multiple_value_parameters:
                    L.info(f"{multi_value.location_str}: {multi_value.values}")

        # Return the multiple_value_parameters
        return self._multiple_value_parameters
    
    @property
    def multiple_value_parameters_dictionary(self, *, display: bool = False) -> dict:
        d = {}
        for multi_value in self.multiple_value_parameters():
            d[multi_value.location_str] = multi_value.values

        return d

    def coordinate_parameters(self, *, display: bool = False) -> list[SingleCoordinateScanParams]:
        """Must be implemented by a subclass of Scan."""
        msg = "coordinate_parameters() must be implemented by a subclass of Scan."
        raise NotImplementedError(msg)

    def coordinate_instances(self, *, display: bool = False) -> list[SingleCoordinateMixin]:
        """Coordinate instance.

        - Returns a list of "coordinate instances" by:
            - Iterating through self.coordinate_parameters()
            - Creating a single "coordinate instance" for each single coordinate parameter

        - Each "coordinate instance" is created by:
            - Making a deep copy of the form
            - Editing the multi value parameters (lists) to have the values of the single
                coordinate parameters
                (i.e. timestamps.timestamps_1.interval = [1.0, 5.0] ->
                    timestamps.timestamps_1.interval = 1.0)
            - Casting the form to its single_coord_class_name type
                (i.e. SimulationsForm -> Simulation)
        """
        self._coordinate_instances = []

        # Iterate through coordinate_parameters
        for idx, single_coordinate_scan_params in enumerate(self.coordinate_parameters()):
            # Make a deep copy of self.form
            single_coordinate_form = copy.deepcopy(self.form)

            # Iterate through parameters in the single_coordinate_parameters tuple
            # Change the value of the multi parameter from a list to the single value of the
            # coordinate
            for scan_param in single_coordinate_scan_params.scan_params:
                level_0_val = single_coordinate_form.__dict__[scan_param.location_list[0]]

                # If the first level is a Block
                if isinstance(level_0_val, Block):
                    level_0_val.__dict__[scan_param.location_list[1]] = scan_param.value

                # If the first level is a category dictionary
                if isinstance(level_0_val, dict):
                    level_1_val = level_0_val[scan_param.location_list[1]]
                    if isinstance(level_1_val, Block):
                        level_1_val.__dict__[scan_param.location_list[2]] = scan_param.value
                    else:
                        msg = f"Non Block parameter {level_1_val} found in Form dictionary: \
                            {level_0_val}"
                        raise TypeError(msg)

            try:
                # Cast the form to its single_coord_class_name type
                coordinate_instance = single_coordinate_form.cast_to_single_coord()

                # Set the variables of the coordinate instance related to the scan
                coordinate_instance.idx = idx
                coordinate_instance.single_coordinate_scan_params = single_coordinate_scan_params

                # Append the coordinate instance to self._coordinate_instances
                self._coordinate_instances.append(coordinate_instance)

            except ValidationError as e:
                raise ValidationError(e) from e

        # Optionally display the coordinate instances
        if display:
            L.info("\nCOORDINATE INSTANCES")
            for coordinate_instance in self._coordinate_instances:
                L.info(coordinate_instance)

        # Return self._coordinate_instances
        return self._coordinate_instances

    def execute(self, processing_method: str = "", data_postprocessing_method: str = "", db_client: entitysdk.client.Client = None):
        """Description."""
        return_dict = {}

        L.info(db_client)

        if not processing_method:
            msg = "Processing method must be specified."
            raise ValueError(msg)
        
        Path.mkdir(self.output_root, parents=True, exist_ok=True)

        # Serialize the scan
        self.serialize(self.output_root / "run_scan_config.json")

        # Create a bbp_workflow_campaign_config
        self.create_bbp_workflow_campaign_config(
            self.output_root / "bbp_workflow_campaign_config.json"
        )

        campaign = None
        if data_postprocessing_method == 'save':
            # Initialize the campaign in the database
            campaign = self.form.initialize_db_campaign(self.output_root, self.multiple_value_parameters_dictionary, db_client)
        
        single_entities = []
        
        # Iterate through self.coordinate_instances()
        for coordinate_instance in self.coordinate_instances():
            # Check if coordinate instance has function "run"
            if hasattr(coordinate_instance, processing_method):
                # Initialize the coordinate_instance's coordinate_output_root
                coordinate_instance.initialize_coordinate_output_root(
                    self.output_root, self.coordinate_directory_option
                )

                # Serialize the coordinate instance
                coordinate_instance.serialize(
                    coordinate_instance.coordinate_output_root / "run_coordinate_instance.json"
                )

                # Call the coordinate_instance's processing_method (i.e. run, generate)
                return_dict[coordinate_instance.idx] = getattr(
                    coordinate_instance, processing_method
                )(db_client)

                # If a data_postprocessing_method is specified, call it
                if data_postprocessing_method:
                    single_entity = return_dict[coordinate_instance.idx] = getattr(
                        coordinate_instance, data_postprocessing_method
                    )(campaign, db_client)
                    single_entities.append(single_entity)

                

            else:
                # Raise an error if run() not implemented for the coordinate instance
                msg = (
                    f"Coordinate instance of type {type(coordinate_instance)} does not have a "
                    f"function {processing_method}."
                )
                raise NotImplementedError(msg)

        if data_postprocessing_method:
            getattr(
                self.form, data_postprocessing_method
            )(single_entities, db_client)

        return campaign
        

    def serialize(self, output_path: Path) -> dict:
        """Serialize a Scan object.

        - type name added to each subobject of type
            inheriting from OBIBaseModel for future deserialization
        """
        # Important to use model_dump_json() instead of model_dump()
        # so OBIBaseModel's custom encoder is used to seri
        # PosixPaths as strings
        model_dump = self.model_dump_json()

        # Now load it back into an ordered dict to do some additional modifications
        model_dump = OrderedDict(json.loads(model_dump))

        # Add the obi_one version to the model_dump
        model_dump["obi_one_version"] = version("obi-one")

        # Order keys in dict
        model_dump.move_to_end("output_root", last=False)
        model_dump.move_to_end("type", last=False)
        model_dump.move_to_end("obi_one_version", last=False)

        # Order the keys in subdict "form"
        model_dump["form"] = OrderedDict(model_dump["form"])
        model_dump["form"].move_to_end("type", last=False)

        # Create the directory and write dict to json file
        if output_path:
            with output_path.open("w") as json_file:
                json.dump(model_dump, json_file, indent=4)

        return model_dump

    def create_bbp_workflow_campaign_config(self, output_path: str = "") -> None:
        """Description."""
        msg = "create_bbp_workflow_campaign_config() not yet complete."
        L.info(msg)
        return

        # Dictionary intialization
        campaign_config = {"dims": [], "attrs": {}, "data": [], "coords": {}, "name": ""}

        # dims
        campaign_config["dims"] = [
            multi_param.location_str for multi_param in self.multiple_value_parameters()
        ]

        multi_value_parameters = self.multiple_value_parameters()
        if len(multi_value_parameters):
            # dims
            campaign_config["dims"] = [
                multi_param.location_str for multi_param in self.multiple_value_parameters()
            ]

            # coords
            for multi_param in multi_value_parameters:
                sub_d = {
                    multi_param.location_str: {
                        "dims": [multi_param.location_str],
                        "attrs": {},
                        "data": multi_param.values,
                    }
                }
                campaign_config["coords"].update(sub_d)
        else:
            campaign_config["dims"] = ["single_coordinate"]
            campaign_config["coords"] = {
                "single_coordinate": {
                    "dims": ["single_coordinate"],
                    "attrs": {},
                    "data": [self.form.single_coord_scan_default_subpath],
                }
            }

        # data
        campaign_config["data"] = [[["a", "b"], ["c", "d"]], [["e", "f"], ["g", "h"]]]

        output_path = Path(output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)

        with output_path.open("w") as json_file:
            json.dump(campaign_config, json_file, indent=4)

    def display_coordinate_parameters(self) -> None:
        """Description."""
        L.info("\nCOORDINATE PARAMETERS")
        for single_coordinate_parameters in self._coordinate_parameters:
            single_coordinate_parameters.display_parameters()

    def save(self) -> None:
        self.form.save(coordinate_instance_entities)

        coordinate_instance_entities = []
        for coordinate_instance in self.coordinate_instances():
            coordinate_instance_entity = coordinate_instance.save()
            coordinate_instance_entities.append(coordinate_instance_entity)

        


class GridScan(Scan):
    """Description."""

    def coordinate_parameters(self, *, display: bool = False) -> list[SingleCoordinateScanParams]:
        """Description."""
        single_values_by_multi_value = []
        multi_value_parameters = self.multiple_value_parameters()
        if len(multi_value_parameters):
            for multi_value in multi_value_parameters:
                single_values = [
                    SingleValueScanParam(location_list=multi_value.location_list, value=value)
                    for value in multi_value.values
                ]

                single_values_by_multi_value.append(single_values)

            self._coordinate_parameters = []
            for scan_params in product(*single_values_by_multi_value):
                self._coordinate_parameters.append(
                    SingleCoordinateScanParams(scan_params=scan_params)
                )

        else:
            self._coordinate_parameters = [
                SingleCoordinateScanParams(
                    nested_coordinate_subpath_str=self.form.single_coord_scan_default_subpath
                )
            ]

        # Optionally display the coordinate parameters
        if display:
            self.display_coordinate_parameters()

        # Return the coordinate parameters
        return self._coordinate_parameters


class CoupledScan(Scan):
    """Description."""

    def coordinate_parameters(self, *, display: bool = False) -> list:
        """Description."""
        previous_len = -1

        multi_value_parameters = self.multiple_value_parameters()
        if len(multi_value_parameters):
            for multi_value in multi_value_parameters:
                current_len = len(multi_value.values)
                if previous_len not in {-1, current_len}:
                    msg = f"Multi value parameters have different lengths: {previous_len} and \
                            {current_len}"
                    raise ValueError(msg)

                previous_len = current_len

            n_coords = current_len

            self._coordinate_parameters = []
            for coord_i in range(n_coords):
                scan_params = [
                    SingleValueScanParam(
                        location_list=multi_value.location_list,
                        value=multi_value.values[coord_i],
                    )
                    for multi_value in multi_value_parameters
                ]
                self._coordinate_parameters.append(
                    SingleCoordinateScanParams(scan_params=scan_params)
                )

        else:
            self._coordinate_parameters = [
                SingleCoordinateScanParams(
                    nested_coordinate_subpath_str=self.form.single_coord_scan_default_subpath
                )
            ]

        if display:
            self.display_coordinate_parameters()

        return self._coordinate_parameters



from obi_one.scientific.simulation.timestamps import (
    SingleTimestamp, 
    RegularTimestamps
)

TimestampsUnion = (
    SingleTimestamp
    | RegularTimestamps
)

from obi_one.core.block_reference import BlockReference
from typing import ClassVar, Any
class TimestampsReference(BlockReference):
    """A reference to a NeuronSet block."""
    
    allowed_block_types: ClassVar[Any] = TimestampsUnion



from obi_one.core.base import OBIBaseModel


class NamedTuple(OBIBaseModel):
    """Helper class to assign a name to a tuple of elements."""

    name: str
    elements: tuple

    def __repr__(self) -> str:
        """Return a string representation of the NamedTuple."""
        return self.name



import os
import warnings

import morphio
import numpy
import pandas
from scipy import stats

try:
    from conntility.subcellular import MorphologyPathDistanceCalculator
except ImportError:
    warnings.warn("Connectome functionalities not available", UserWarning, stacklevel=1)


def morphology_and_pathdistance_calculator(circ, node_population, node_id):
    """Loads for a specified neuron its morphology and creates a path-distance calculator object.

    Args:
      circ (bluepysnap.Circuit): The circuit the neuron resides in
      node_population (str): Name of the node population of the neuron
      node_id (int): Neuron node id.
    """
    node = circ.nodes[node_population]
    morph_name = node.morph.get_name(node_id)
    try:
        morph = morphio.Morphology(
            os.path.join(node.config["alternate_morphologies"]["h5v1"], morph_name) + ".h5"
        )
    except:
        raise RuntimeError(f"Error loading hdf5 morphology for {node_population} - {node_id}")
    PD = MorphologyPathDistanceCalculator(morph)
    return morph, PD


def all_syns_on(circ, node_population, node_id, node_props):
    """Load for a specified neuron relevant properties of all its afferent synapses across
    edge populations.

    Args:
      circ (bluepysnap.Circuit): The circuit the neuron resides in
      node_population (str): Name of the node population of the neuron
      node_id (int): Neuron node id.
      node_props (list): List of node properties. The values of these properties for
        the presynaptic partners of the afferent synapses will be loaded.
    """
    node = circ.nodes[node_population]
    syns = []
    syn_edge_names = []
    syn_pre_pop_names = []
    pre_node_props = []

    syn_props = [
        "afferent_section_id",
        "afferent_segment_id",
        "afferent_segment_offset",
        "@source_node",
        "@target_node",
    ]
    reserved_props = syn_props + ["source_population", "edge_population", "edge_id"]
    int_props = [
        "afferent_section_id",
        "afferent_segment_id",
        "@source_node",
        "@target_node",
        "edge_id",
    ]
    _node_props = [_prop for _prop in node_props if _prop not in reserved_props]

    for edge_name in node.target_in_edges():
        edge = circ.edges[edge_name]
        if not numpy.all([_x in edge.property_names for _x in syn_props]):
            print("Skipping!")
            continue
        new_syns = edge.afferent_edges(node_id, properties=syn_props)
        if len(new_syns) == 0:
            continue
        syns.append(new_syns)
        pre_node = circ.nodes[edge.source.name]
        __node_props = [_x for _x in _node_props if _x in pre_node.property_names]
        loaded_node_props = pre_node.get(syns[-1]["@source_node"], properties=__node_props)
        loaded_node_props = loaded_node_props.reindex(
            columns=_node_props, index=syns[-1]["@source_node"].values
        )
        pre_node_props.append(loaded_node_props)
        syn_edge_names.append(edge_name)
        syn_pre_pop_names.append(edge.source.name)

    syns = pandas.concat(
        syns,
        axis=0,
        names=["source_population", "edge_population"],
        keys=list(zip(syn_pre_pop_names, syn_edge_names, strict=False)),
    )
    syns = syns.reset_index([2]).rename(columns={"level_2": "edge_id"})
    pre_node_props = pandas.concat(
        pre_node_props,
        axis=0,
        names=["source_population", "edge_population"],
        keys=list(zip(syn_pre_pop_names, syn_edge_names, strict=False)),
    )
    pre_node_props = pre_node_props.reset_index([2]).drop(columns=["node_ids"])

    syns = pandas.concat([syns, pre_node_props], axis=1).reset_index()

    for prop in int_props:
        syns[prop] = syns[prop].astype(int)

    return syns


# Section types are also in the synapse table, but I do not trust its presence in all circuits
def add_section_types(syns, morph):
    """Load the section types of synaptic locations on a morphology.

    Args:
      syns (pandas.DataFrame): Dataframe that includes morphology locations in its columns, as
        they are specified in SONATA, e.g, "afferent_section_id".
      morph (morphio.Morphology): The morphology the locations are defined on
    """
    sec_types = numpy.array([1] + [int(_sec.type) for _sec in morph.sections])
    syns["afferent_section_type"] = sec_types[syns["afferent_section_id"]]


def apply_filters(syns, filter_dict, drop_nan=True):
    """Filters a dataframe of synaptic locations according to specified property values.

    Args:
      syns (pandas.DataFrame): DataFrame with synapse properties. Can really be any types
        of properties as long as they are in different columns.
      filter_dict (dict): Specified filters. Keys of the dict specify the names of columns
        of the dataframe, values the values that pass the filter. A row has to pass ALL
        filters to be kept.
      drop_nan (bool, default: True): If set to False, then a nan value in a row is considered
        to pass the filter. Else, the row is dropped.
        Note however that nan values in columns that are not named in the keys of the
        filter_dict are ignored.
    """
    for k, v in filter_dict.items():
        assert k in syns.columns, (
            f"No property {k} could be loaded"
        )  # This cannot happen! We check for it earlier.
        if isinstance(v, list):
            v = syns[k].isin(v)
        else:
            v = syns[k] == v
        if not drop_nan:
            v = v | syns[k].isna()
        syns = syns.loc[v]
    return syns


def relevant_path_distances(PD, syns):
    """Calculates and return path distances to the soma and all pairwise path distances for
    dendritic locations in a dataframe.

    Args:
      PD (conntility.subcellular.PathDistanceCalculator): Calculator for the relevant morphology.
      syns (pandas.DataFrame): Dataframe that includes morphology locations in its columns, as
        they are specified in SONATA, e.g, "afferent_section_id".
    """
    pw_pds = PD.path_distances(syns)

    soma = pandas.DataFrame(
        {"afferent_section_id": [0], "afferent_segment_id": [0], "afferent_segment_offset": [0.0]}
    )
    soma_pds = PD.path_distances(soma, locs_to=syns)[0]
    return soma_pds, pw_pds


def select_randomly(syns, n=None, p=None, raise_insufficient=False):
    """From a set of synapse locations, given in a DataFrame, select some of them randomly.

    Args:
      syns (pandas.DataFrame): DataFrame with synapse properties. Can really be any types
        of properties as long as they are in different columns.
      n (int): Specify the number of synapses (i.e., rows of syns) to return.
      p (float): Specify that probability that each given synapse (i.e., row of syns) is
        selected and returned. If n is provided, this is ignored.
      raise_insufficient (bool, default=False): If set to True, then if n > len(syns) an
        exception is raised.
    """
    assert p is not None or n is not None, "Must specify number or fraction of synapses!"
    if n is not None:
        if n > len(syns):
            if raise_insufficient:
                raise RuntimeError(f"Fewer than the requested count of {n} found!")
        return syns.iloc[numpy.random.choice(len(syns), numpy.minimum(n, len(syns)), replace=False)]
    if p is not None:
        if p < 0.0 or p > 1.0:
            raise ValueError("p must be between 0 and 1!")
        picked = numpy.random.rand(len(syns)) < p
        return syns.loc[picked]


def select_minmax_distance(
    syns, soma_pds, soma_pd_min, soma_pd_max, n=None, p=None, raise_insufficient=False
):
    """From a set of synapse locations, given in a DataFrame, select some of them randomly.
    But only select locations that are between a specified minimum and maximum path distance
    to the soma.

    Args:
      syns (pandas.DataFrame): DataFrame with synapse properties. Can really be any types
        of properties as long as they are in different columns.
      soma_pds (numpy.array): Specifies for each synapse its soma path distance.
      soma_pd_min (float): minimum soma path distance admitted.
      soma_pd_max (float): Maximum soma path distance admitted.
      n (int): Specify the number of synapses (i.e., rows of syns) to return.
      p (float): Specify that probability that each given synapse (i.e., row of syns) is
        selected and returned. If n is provided, this is ignored.
      raise_insufficient (bool, default=False): If set to True, then if n is larger than the
        number of synapses within the path distance interval, an exception is raised.
    """
    valid = (soma_pds >= soma_pd_min) & (soma_pds < soma_pd_max)
    return select_randomly(syns.loc[valid], n=n, p=p, raise_insufficient=raise_insufficient)


def select_closest_to_path_distance(syns, soma_pds, target_soma_pd, n, raise_insufficient=False):
    """From a set of synapse locations, given in a DataFrame, select the ones that are closest
    to a specified path distance from the soma.

    Args:
      syns (pandas.DataFrame): DataFrame with synapse properties. Can really be any types
        of properties as long as they are in different columns.
      soma_pds (numpy.array): Specifies for each synapse its soma path distance.
      target_soma_pd (float): Soma path distance that you want to get close to.
      n (int): Specify the number of synapses (i.e., rows of syns) to return. The
        n closest to the specified path distance will be picked. Note that they can
        be on very different branches.
      raise_insufficient (bool, default=False): If set to True, then if n is larger than
       len(syns) an exception is raised.
    """
    if n > len(syns):
        if raise_insufficient:
            raise RuntimeError(f"Fewer than the requested count of {n} found!")
    srt_idx = numpy.argsort(numpy.abs(soma_pds - target_soma_pd))
    return syns.iloc[srt_idx[:n]]


def _pd_gaussian_selector(soma_pds, soma_pd_mean, soma_pd_sd, n, raise_insufficient=False):
    """From a list of soma path distances, select some of them randomly with probabilities
    that depend on values of a path distance-dependent Gaussian.
    Returns the indices of elements that are picked from the list of path distances.

    Args:
      soma_pds (numpy.array): An array of path distance values.
      soma_pd_mean (float): Mean of a Gaussian that defines the probability function.
      soma_pd_sd (float): SD of the same Gaussian
      n (int): Specify the number of synapses (i.e., rows of syns) to return.
      raise_insufficient (bool, default=False): If set to True, then if n > len(syns) an
        exception is raised.
    """
    if (n > len(soma_pds)) and raise_insufficient:
        raise RuntimeError(f"Fewer than the requested count of {n} found!")

    distr = stats.norm(soma_pd_mean, soma_pd_sd).pdf(soma_pds)
    sel_idx = numpy.random.choice(
        range(len(soma_pds)), numpy.minimum(n, len(soma_pds)), p=distr / distr.sum()
    )
    return sel_idx


def select_by_path_distance(
    syns, soma_pds, soma_pd_mean, soma_pd_sd, n=None, p=None, raise_insufficient=False
):
    """From a set of synapse locations, given in a DataFrame, select some of them randomly.
    The relative probabilities that each synapse is picked are defined by a parameterized
    Gaussian that depends on the path distances to the soma of the synapses.

    Args:
      syns (pandas.DataFrame): DataFrame with synapse properties. Can really be any types
        of properties as long as they are in different columns.
      soma_pds (numpy.array): An array of path distance values.
      soma_pd_mean (float): Mean of a Gaussian that defines the probability function.
      soma_pd_sd (float): SD of the same Gaussian
      n (int): Specify the number of synapses (i.e., rows of syns) to return.
      p (float): Specify that probability that each given synapse (i.e., row of syns) is
        selected and returned. If n is provided, this is ignored.
      raise_insufficient (bool, default=False): If set to True, then if n > len(syns) an
        exception is raised.
    """
    assert p is not None or n is not None, "Must specify number or fraction of synapses!"
    if p is not None:
        if p < 0.0 or p > 1.0:
            raise ValueError("p must be between 0 and 1!")
        n = stats.binom(len(syns), p).rvs()
    return syns.iloc[
        _pd_gaussian_selector(
            soma_pds, soma_pd_mean, soma_pd_sd, n, raise_insufficient=raise_insufficient
        )
    ]


def select_clusters_by_max_distance(
    syns,
    soma_pds,
    pw_pds,
    n_clusters,
    cluster_max_distance,
    soma_pd_mean=None,
    soma_pd_sd=None,
    raise_insufficient=False,
):
    """From a set of synapse locations, given in a DataFrame, select some of them randomly.
    The selected synapses will be clustered in the sense that their pairwise path distances
    are below a given value.
    Optionally, pick probabilities of all synapses depend on their soma path distance.

    Args:
      syns (pandas.DataFrame): DataFrame with synapse properties. Can really be any types
        of properties as long as they are in different columns.
      soma_pds (numpy.array): An array of path distance values for all synapses.
      pw_pds (numpy.array, len(syns) X len(syns)): Pairwise path distances between all pairs
      of locations in syns.
      n_clusters (int): Number of clusters to generate.
      cluster_max_distance (float): The maximum path distance of synapses within a cluster
      from the center of that cluster. All synapses within that distance will be picked.
      soma_pd_mean (float, optional): Mean of a Gaussian that defines a path distance-dependent
        probability function that biases the synapse selection. If it is not specified, then
        synapses at all path distances are equally likely to be picked.
      soma_pd_sd (float, optional): SD of the same Gaussian.
      raise_insufficient (bool, default=False): If set to True, then if the specified number
      of clusters cannot be generated, an exception is raised.
    """
    syns_out = []
    for _ in range(n_clusters):
        if len(syns) == 0:
            if raise_insufficient:
                raise RuntimeError(
                    f"Fewer than the requested count of {n_clusters} clusters possible!"
                )
            break
        if soma_pd_mean is not None and soma_pd_sd is not None:
            ctr = _pd_gaussian_selector(
                soma_pds, soma_pd_mean, soma_pd_sd, 1, raise_insufficient=True
            )[0]
        else:
            ctr = numpy.random.choice(len(syns))
        clstr_ids = pw_pds[ctr] < cluster_max_distance
        syns_out.append(syns.loc[clstr_ids])
        syns = syns.loc[~clstr_ids]
        pw_pds = pw_pds[numpy.ix_(~clstr_ids, ~clstr_ids)]
        if soma_pds is not None:
            soma_pds = soma_pds[~clstr_ids]
    return pandas.concat(
        syns_out, axis=0, names=["cluster_id"], keys=range(len(syns_out))
    ).reset_index(0)


def select_clusters_by_count(
    syns,
    soma_pds,
    pw_pds,
    n_clusters,
    n_per_cluster,
    soma_pd_mean=None,
    soma_pd_sd=None,
    raise_insufficient=False,
):
    """From a set of synapse locations, given in a DataFrame, select some of them randomly.
    The selected synapses will be clustered in the sense that synapses are selected
    together with all their nearest neighbors in terms of path distance.
    Optionally, pick probabilities of all synapses depend on their soma path distance.

    Args:
      syns (pandas.DataFrame): DataFrame with synapse properties. Can really be any types
        of properties as long as they are in different columns.
      soma_pds (numpy.array): An array of path distance values for all synapses.
      pw_pds (numpy.array, len(syns) X len(syns)): Pairwise path distances between all pairs
      of locations in syns.
      n_clusters (int): Number of clusters to generate.
      n_per_cluster (int): After picking a synapse as the center of a cluster, its
      n_per_cluster - 1 nearest neighbors are also included. However, synapses that have been
      picked for one cluster cannot be picked again for another cluster, even if sufficiently
      near.
      soma_pd_mean (float, optional): Mean of a Gaussian that defines a path distance-dependent
        probability function that biases the synapse selection. If it is not specified, then
        synapses at all path distances are equally likely to be picked.
      soma_pd_sd (float, optional): SD of the same Gaussian.
      raise_insufficient (bool, default=False): If set to True, then if the specified number
      of clusters cannot be generated, an exception is raised.
    """
    syns_out = []
    for _ in range(n_clusters):
        if len(syns) < n_per_cluster:
            if raise_insufficient:
                raise RuntimeError(
                    f"Fewer than the requested count of {n_clusters} clusters possible!"
                )
            break
        if soma_pd_mean is not None and soma_pd_sd is not None:
            ctr = _pd_gaussian_selector(
                soma_pds, soma_pd_mean, soma_pd_sd, 1, raise_insufficient=True
            )[0]
        else:
            ctr = numpy.random.choice(len(syns))
        clstr_ids = numpy.argsort(pw_pds[ctr])[:n_per_cluster]
        other_ids = numpy.setdiff1d(range(len(syns)), clstr_ids)
        syns_out.append(syns.iloc[clstr_ids])
        syns = syns.iloc[other_ids]
        pw_pds = pw_pds[numpy.ix_(other_ids, other_ids)]
        if soma_pds is not None:
            soma_pds = soma_pds[other_ids]
    return pandas.concat(
        syns_out, axis=0, names=["cluster_id"], keys=range(len(syns_out))
    ).reset_index(0)


def merge_multiple_syns_per_connection(syns, soma_pds, pw_pds):
    syns = syns.reset_index(drop=True)
    _grp = (
        syns.reset_index(drop=False)
        .groupby(["source_population", "@source_node"])["index"]
        .apply(list)
    )
    mn_soma_pd = _grp.apply(lambda _x: soma_pds[_x].mean())
    mn_pw_pds = _grp.apply(lambda _x: _grp.apply(lambda _y: pw_pds[numpy.ix_(_x, _y)].mean()))

    syns_out = mn_soma_pd.index.to_frame().reset_index(drop=True)
    mn_soma_pd = mn_soma_pd.values
    mn_pw_pds = mn_pw_pds.values
    return syns_out, mn_soma_pd, mn_pw_pds



from abc import ABC, abstractmethod
from typing import Annotated, Self, ClassVar

from pydantic import Field, NonNegativeFloat, PositiveFloat, model_validator

from obi_one.core.block import Block
from obi_one.scientific.unions.unions_neuron_sets import NeuronSetUnion, NeuronSetReference
from obi_one.scientific.circuit.circuit import Circuit
from obi_one.core.exception import OBIONE_Error


class Recording(Block, ABC):

    neuron_set: Annotated[NeuronSetReference, Field(title="Neuron Set", description="Neuron set to record from.")]

    _start_time: NonNegativeFloat = 0.0
    _end_time: PositiveFloat = 100.0

    dt: Annotated[
        PositiveFloat | list[PositiveFloat],
        Field(default=0.1,
            title="Timestep",
            description="Interval between recording time steps in milliseconds (ms).", units="ms"),
    ]

    def config(self, circuit: Circuit, population: str | None=None, end_time: NonNegativeFloat | None = None) -> dict:
        self.check_simulation_init()

        if self.neuron_set.block.population_type(circuit, population) != "biophysical":
            raise OBIONE_Error(
                f"Neuron Set '{self.neuron_set.block.name}' for {self.__class__.__name__}: \'{self.name}\' should be biophysical!"
            )
        
        if end_time is None:
            raise OBIONE_Error(f"End time must be specified for recording '{self.name}'.")
        self._end_time = end_time

        sonata_config = self._generate_config()

        if self._end_time <= self._start_time:
            raise OBIONE_Error(
                f"Recording '{self.name}' for Neuron Set '{self.neuron_set.block.name}': "
                "End time must be later than start time!"
            )
        
        return sonata_config

    @abstractmethod
    def _generate_config(self) -> dict:
        pass


class SomaVoltageRecording(Recording):
    """Records the soma voltage of a neuron set for the full length of the experiment."""

    title: ClassVar[str] = "Soma Voltage Recording (Full Experiment)"

    def _generate_config(self) -> dict:
        sonata_config = {}

        sonata_config[self.name] = {
            "cells": self.neuron_set.block.name,
            "sections": "soma",
            "type": "compartment",
            "compartments": "center",
            "variable_name": "v",
            "unit": "mV",
            "dt": self.dt,
            "start_time": self._start_time,
            "end_time": self._end_time,
        }
        return sonata_config
    

class TimeWindowSomaVoltageRecording(SomaVoltageRecording):
    """Records the soma voltage of a neuron set over a specified time window."""

    title: ClassVar[str] = "Soma Voltage Recording (Time Window)"

    start_time: Annotated[
        NonNegativeFloat | list[NonNegativeFloat], Field(default=0.0, description="Recording start time in milliseconds (ms).", units="ms")
    ]
    end_time: Annotated[
        NonNegativeFloat | list[NonNegativeFloat], Field(default=100.0, description="Recording end time in milliseconds (ms).", units="ms")
    ]

    @model_validator(mode="after")
    def check_start_end_time(self) -> Self:
        """Check that end time is later than start time."""
        if self.end_time <= self.start_time:
            if self.has_name():
                recording_name = f" '{self.name}'"
            else:
                recording_name = ""

            if self.neuron_set.has_block() and self.neuron_set.block.has_name():
                neuron_set_name = f" '{self.neuron_set.block.name}'"
            else:
                neuron_set_name = ""

            raise OBIONE_Error(
                f"Recording{recording_name} for Neuron Set{neuron_set_name}: "
                "End time must be later than start time!"
            )
        return self

    def _generate_config(self) -> dict:

        self._start_time = self.start_time
        self._end_time = self.end_time
        
        return super()._generate_config()



from obi_one.core.block import Block


class ExtracellularLocationSet(Block):
    """ """


class XYZExtracellularLocationSet(ExtracellularLocationSet):
    xyz_locations: (
        tuple[tuple[float, float, float], ...] | list[tuple[tuple[float, float, float], ...]]
    ) = ((0.0, 0.0, 0.0),)



from typing import ClassVar
from pydantic import model_validator

from obi_one.core.base import OBIBaseModel
from obi_one.core.block import Block
from obi_one.core.block_reference import BlockReference

from typing import get_args, get_origin

import types



class Form(OBIBaseModel):
    """A Form is a configuration for single or multi-dimensional parameter scans.

    A Form is composed of Blocks, which either appear at the root level
    or within dictionaries of Blocks where the dictionary is takes a Union of Block types.
    """

    name: ClassVar[str] = "Add a name class' name variable"
    description: ClassVar[str] = """Add a description to the class' description variable"""
    single_coord_class_name: ClassVar[str] = ""

    _block_mapping: dict = None

    @classmethod
    def empty_config(cls):
        # Here you use model_construct or build custom behavior
        return cls.model_construct()
    
    def validated_config(self):
        return self.__class__.model_validate(self.model_dump())
    

    @property
    def block_mapping(self) -> dict:
        """Returns a mapping of block class names to block_dict_name and reference_type."""

        if self._block_mapping is None:
            # Get type annotations of the instance's class
            annotations = self.__class__.__annotations__

            self._block_mapping = {}
            for attr_name, attr_value in self.__dict__.items():
                # Check if it's a dictionary of Block instances
                if isinstance(attr_value, dict) and all(
                    isinstance(v, Block) for v in attr_value.values()
                ):
                    # Get the annotated type of this attribute (e.g., dict[str, TimestampsUnion])
                    annotated_type = annotations.get(attr_name)

                    field_info = self.__pydantic_fields__[attr_name]
                    reference_type = None

                    if field_info.json_schema_extra and "reference_type" in field_info.json_schema_extra:
                        reference_type = field_info.json_schema_extra["reference_type"]
                    else: 
                        raise ValueError(
                            f"Attribute '{attr_name}' does not have a 'reference_type' in json_schema_extra."
                        )

                    if annotated_type and get_origin(annotated_type) is dict:
                        value_type = get_args(annotated_type)[1]  # dict[key_type, value_type]

                        if hasattr(types, "UnionType") and isinstance(value_type, types.UnionType):
                            classes = [arg for arg in get_args(value_type)]
                        else:
                            classes = [value_type]

                    for block_class in classes:
                        if block_class.__name__ in self._block_mapping:
                            # If the block class is already in the mapping, append the new block
                            raise ValueError(
                                f"Block class {block_class.__name__} already exists in the mapping. \
                                    This suggests that the same block class is used in multiple dictionaries."
                            )
                        else:
                            # Initialize a new dictionary for this block class
                            self._block_mapping[block_class.__name__] = {
                                "block_dict_name": attr_name,
                                "reference_type": reference_type,
                            }

        return self._block_mapping



    

    def fill_block_reference_for_block(self, block: Block):
        """Fill the block reference with the actual Block object it references."""

        for block_attr_name, block_attr_value in block.__dict__.items():
            # If the Block instance has a `BlockReference` attribute, set it the object it references
            if isinstance(block_attr_value, BlockReference):
                block_reference = block_attr_value

                if block_reference.block_dict_name != "" and block_reference.block_name != "":
                    block_reference.block = self.__dict__[block_reference.block_dict_name][block_reference.block_name]
                elif block_reference.block_dict_name == "" and block_reference.block_name != "":
                    # If the block_dict_name is empty, we assume the block_name is a direct reference to a Block instance
                    if block_reference.block_name == "neuron_set_extra":
                        block_reference.block = self.__dict__[block_reference.block_name]
                else:
                    raise ValueError("BlockReference must have a non-empty block_dict_name and block_name.")
        

    @model_validator(mode="after")
    def fill_block_references(self):

        for attr_name, attr_value in self.__dict__.items():
            # Check if the attribute is a dictionary of Block instances
            if isinstance(attr_value, dict) and all(
                isinstance(dict_val, Block) for dict_key, dict_val in attr_value.items()
            ):
                category_blocks_dict = attr_value

                # If so iterate through the dictionary's Block instances
                for _, block in category_blocks_dict.items():
                    self.fill_block_reference_for_block(block)
            
            elif isinstance(attr_value, Block):
                block = attr_value
                self.fill_block_reference_for_block(block)

        return self


    def cast_to_single_coord(self) -> OBIBaseModel:
        """Cast the form to a single coordinate object."""
        module = __import__(self.__module__)
        class_to_cast_to = getattr(module, self.single_coord_class_name)
        single_coord = class_to_cast_to.model_construct(**self.__dict__)
        single_coord.type = self.single_coord_class_name
        return single_coord

    @property
    def single_coord_scan_default_subpath(self) -> str:
        return self.single_coord_class_name + "/"



from .morphology_location_block import (
    ClusteredGroupedMorphologyLocations as ClusteredGroupedMorphologyLocations,
    ClusteredMorphologyLocations as ClusteredMorphologyLocations,
    ClusteredPathDistanceMorphologyLocations as ClusteredPathDistanceMorphologyLocations,
    MorphologyLocationsBlock as MorphologyLocationsBlock,
    PathDistanceMorphologyLocations as PathDistanceMorphologyLocations,
    RandomGroupedMorphologyLocations as RandomGroupedMorphologyLocations,
    RandomMorphologyLocations as RandomMorphologyLocations,
)
from .morphology_location_form import (
    MorphologyLocations as MorphologyLocations,
    MorphologyLocationsForm as MorphologyLocationsForm,
)



from obi_one.scientific.basic_connectivity_plots.basic_connectivity_plots import (
    BasicConnectivityPlots,
)
from obi_one.scientific.circuit_extraction.circuit_extraction import (
    CircuitExtractions,
)
from obi_one.scientific.connectivity_matrix_extraction.connectivity_matrix_extraction import (
    ConnectivityMatrixExtractions,
)
from obi_one.scientific.folder_compression.folder_compression import (
    FolderCompressions,
)
from obi_one.scientific.morphology_containerization.morphology_containerization import (
    MorphologyContainerizationsForm,
)
from obi_one.scientific.morphology_containerization.morphology_decontainerization import (
    MorphologyDecontainerizationsForm,
)
from obi_one.scientific.morphology_locations import MorphologyLocationsForm
from obi_one.scientific.morphology_metrics.morphology_metrics import (
    MorphologyMetricsForm,
)
from obi_one.scientific.simulation.simulations import SimulationsForm
from obi_one.scientific.test_forms.test_form_single_block import (
    MultiBlockEntitySDKTestForm,
    SingleBlockEntityTestForm,
    SingleBlockGenerateTestForm,
)

FormUnion = (
    BasicConnectivityPlots
    | CircuitExtractions
    | ConnectivityMatrixExtractions
    | FolderCompressions
    | MorphologyContainerizationsForm
    | MorphologyDecontainerizationsForm
    | MorphologyMetricsForm
    | SimulationsForm
    | SingleBlockGenerateTestForm
    | SingleBlockEntityTestForm
    | MultiBlockEntitySDKTestForm
    | MorphologyLocationsForm
)



from obi_one.core.block import Block
from obi_one.core.base import OBIBaseModel


from typing import Union, ClassVar, get_args, Any, Literal
from pydantic import Field
import abc
class BlockReference(OBIBaseModel, abc.ABC):
    block_dict_name: str = Field(default="")
    block_name: str = Field()

    allowed_block_types: ClassVar[Any] = None
    
    _block: Any = None

    @classmethod
    def allowed_block_type_names(cls, allowed_block_types):
        if allowed_block_types is None:
            return []
        return [t.__name__ for t in get_args(allowed_block_types)]

    class Config:
        @staticmethod
        def json_schema_extra(schema: dict, model) -> None:
            # Dynamically get allowed_block_types from subclass
            allowed_block_types = getattr(model, 'allowed_block_types', [])
            schema['allowed_block_types'] = [t.__name__ for t in get_args(allowed_block_types)]
            schema['is_block_reference'] = True

    @property
    def block(self) -> Block:
        """Returns the block associated with this reference."""
        if self._block is None:
            raise ValueError("Block has not been set.")
        return self._block

    def has_block(self) -> bool:
        return self._block is not None

    @block.setter
    def block(self, value: Block) -> None:
        """Sets the block associated with this reference."""
        if not isinstance(value, self.allowed_block_types):
            raise TypeError(f"Value must be of type {self.block_type.__name__}.")
        self._block = value




from obi_one.scientific.simulation.recording import SomaVoltageRecording, TimeWindowSomaVoltageRecording

RecordingUnion = (SomaVoltageRecording |
                  TimeWindowSomaVoltageRecording)

from obi_one.core.block_reference import BlockReference
from typing import ClassVar, Any
class RecordingReference(BlockReference):
    """A reference to a StimulusUnion block."""
    
    allowed_block_types: ClassVar[Any] = RecordingUnion


from obi_one.scientific.simulation.manipulations import (
    SynapticMgManipulation, 
    ScaleAcetylcholineUSESynapticManipulation
)

SynapticManipulationsUnion = (
    SynapticMgManipulation
    | ScaleAcetylcholineUSESynapticManipulation
)

from obi_one.core.block_reference import BlockReference
from typing import ClassVar, Any
class SynapticManipulationsReference(BlockReference):
    """A reference to a SynapticManipulations block."""

    allowed_block_types: ClassVar[Any] = SynapticManipulationsUnion



from obi_one.scientific.afferent_synapse_finder.specified_afferent_synapses_block import (
    AfferentSynapsesBlock,
    ClusteredPDSynapsesByCount,
    ClusteredPDSynapsesByMaxDistance,
    ClusteredSynapsesByCount,
    ClusteredSynapsesByMaxDistance,
    PathDistanceConstrainedFractionOfSynapses,
    PathDistanceConstrainedNumberOfSynapses,
    PathDistanceWeightedFractionOfSynapses,
    PathDistanceWeightedNumberOfSynapses,
    RandomlySelectedFractionOfSynapses,
    RandomlySelectedNumberOfSynapses,
)

SynapseSetUnion = (
    AfferentSynapsesBlock
    | RandomlySelectedFractionOfSynapses
    | RandomlySelectedNumberOfSynapses
    | PathDistanceConstrainedFractionOfSynapses
    | PathDistanceConstrainedNumberOfSynapses
    | PathDistanceWeightedFractionOfSynapses
    | PathDistanceWeightedNumberOfSynapses
    | ClusteredPDSynapsesByCount
    | ClusteredPDSynapsesByMaxDistance
    | ClusteredSynapsesByCount
    | ClusteredSynapsesByMaxDistance
)



import logging
import os
import warnings
from typing import ClassVar

from bluepysnap import Circuit

L = logging.getLogger(__name__)

from obi_one.core.block import Block
from obi_one.core.form import Form
from obi_one.core.path import NamedPath
from obi_one.core.single import SingleCoordinateMixin

try:
    from conntility.connectivity import ConnectivityMatrix
except ImportError:
    warnings.warn("Connectome functionalities not available", UserWarning, stacklevel=1)


class ConnectivityMatrixExtractions(Form):
    """ """

    single_coord_class_name: ClassVar[str] = "ConnectivityMatrixExtraction"
    name: ClassVar[str] = "Connectivity Matrix Extraction"
    description: ClassVar[str] = (
        "Extracts a connectivity matrix of a given edge population of a SONATA circuit in ConnectomeUtilities format, consisting of a sparse connectivity matrix with the number of synapses for each connection, together with a table (dataframe) of selected node attributes."
    )

    class Initialize(Block):
        circuit_path: NamedPath | list[NamedPath]
        edge_population: None | str | list[None | str] = None
        node_attributes: None | tuple[str, ...] | list[None | tuple[str, ...]] = None

    initialize: Initialize


class ConnectivityMatrixExtraction(ConnectivityMatrixExtractions, SingleCoordinateMixin):
    """Extracts a connectivity matrix of a given edge population of a SONATA circuit in ConnectomeUtilities format,
    consisting of a sparse connectivity matrix with the number of synapses for each connection, together with a
    table (dataframe) of selected node attributes.
    """

    DEFAULT_ATTRIBUTES: ClassVar[tuple[str, ...]] = (
        "x",
        "y",
        "z",
        "mtype",
        "etype",
        "layer",
        "synapse_class",
    )

    def run(self) -> None:
        try:
            L.info(f"Info: Running idx {self.idx}")

            output_file = os.path.join(self.coordinate_output_root, "connectivity_matrix.h5")
            assert not os.path.exists(output_file), f"Output file '{output_file}' already exists!"

            # Load circuit
            L.info(f"Info: Loading circuit '{self.initialize.circuit_path}'")
            c = Circuit(self.initialize.circuit_path.path)
            popul_names = c.edges.population_names
            assert len(popul_names) > 0, "Circuit does not have any edge populations!"
            edge_popul = self.initialize.edge_population
            if edge_popul is None:
                assert len(popul_names) == 1, (
                    "Multiple edge populations found - please specify name of edge population 'edge_popul' to extract connectivity from!"
                )
                edge_popul = popul_names[0]  # Selecting the only one
            else:
                assert edge_popul in popul_names, (
                    f"Edge population '{edge_popul}' not found in circuit!"
                )

            # Extract connectivity matrix
            if self.initialize.node_attributes is None:
                node_props = self.DEFAULT_ATTRIBUTES
            else:
                node_props = self.initialize.node_attributes
            load_cfg = {
                "loading": {
                    "properties": node_props,
                }
            }
            L.info(f"Node properties to extract: {node_props}")
            L.info(f"Extracting connectivity from edge population '{edge_popul}'")
            dummy_edge_prop = next(
                filter(lambda x: "@" not in x, c.edges[edge_popul].property_names)
            )  # Select any existing edge property w/o "@"
            cmat = ConnectivityMatrix.from_bluepy(
                c, load_cfg, connectome=edge_popul, edge_property=dummy_edge_prop, agg_func=len
            )
            # Note: edge_property=<any property> and agg_func=len required to obtain the number of synapses per connection

            # Save to file
            cmat.to_h5(output_file)
            if os.path.exists(output_file):
                L.info(f"Connectivity matrix successfully written to '{output_file}'")

        except Exception as e:
            L.error(f"Error: {e}")



"""
Helper function to extract nodes in simplices in a matrix
Author: Daniela Egas Santander 
Last updated: 06.2024
"""

import numpy as np
import pandas as pd
from conntility import ConnectivityMatrix
from connalysis.network.topology import list_simplices_by_dimension, node_participation

def simplex_submat(adj, v, dim, v_position="source", subsample=False, n_count_max=None,  subsample_method="node_participation", simplex_type='directed', seed=None): 
    """
    Extracts the indices of nodes in the adjacency matrix that participate in simplices of dimension dim with v as a source or target

    Parameters
    ----------
    adj : scipy sparse matrix
        Adjacency matrix of the network (square, will be converted to binary if not already binary).
    v : int
        Index of the node of interest (source or target).
    dim : int
        Dimension of the simplex to extract.
    v_position : {'source', 'target'}, optional
        Whether the node `v` is considered as a source or target in the simplices (default: 'source').
    subsample : bool, optional 
        Whether to subsample the nodes if there are too many (default: False).
    n_count_max : int, optional
        Maximum number of nodes to return if subsampling (required if subsample=True).
    subsample_method : {'node_participation', 'random', 'sample_simplices'}, optional
        Method for subsampling nodes if needed (default: 'node_participation').
    simplex_type : {'directed', 'reciprocal', 'undirected'}, optional
        Type of simplex to extract (default: 'directed').
    seed : int, optional
        Random seed for reproducibility for random subsampling.

    Returns
    -------
    np.ndarray or tuple of np.ndarrays
        If `nodes` is False, returns the nodes in simplices of dimension `dim` with `v` as source or target
        If `subsample` is True, returns a tuple (`selection`, `nodes`), where `nodes` is as above and 
        `selection` is the subsampled set.

    Notes
    -----
    - Subsampling methods:
        - 'random': randomly selects nodes from all nodes.
        - 'node_participation': selects nodes with highest node participation in dimension dim in the submatrix on the nodes in `nodes`.
        - 'sample_simplices': samples simplices while the number of nodes on them is still smaller or equal than the required size.
    - If the number of candidate nodes is less than or equal to `n_count_max`, no subsampling is performed.
    - If the number of sampled nodes is smaller than the nodes in a simplex of dimension `dim` there is no possible solution.

    """
    # Basic checks 
    assert adj.shape[0] == adj.shape[1], "Adjacency matrix must be square"
    assert 0 <= v < adj.shape[0], f"v must be between 0 and {adj.shape[0]-1} since its a node in adj"
    if subsample:
        assert isinstance(n_count_max, int), "n_count_max must be an integer when subsampling"
    if v_position=="target": 
        adj=adj.transpose()
    adj=adj.astype(bool).astype(int).tocsr()

    # Get simplex list on v 
    sl=list_simplices_by_dimension(adj,max_dim=dim,nodes=np.array([v]),simplex_type=simplex_type)
    # Check if dimension and n_count_max are valid
    if dim > sl.index.max(): 
        dim=sl.index.max()
        print(f"> Dimension not attained using dimension {dim} instead.")
    if subsample and (n_count_max < dim + 1):
        n_count_max = dim+1
        print(f"> n_count_max is too small to form a single {dim}-simplex, sampling n_count_max = {n_count_max} neurons instead.")
        
    # Get nodes 
    selection_test=np.unique(sl.loc[dim])
    if not subsample: 
        return selection_test
    else:
        if selection_test.shape[0]<=n_count_max:
            print("No subselection required")
            selection=selection_test
        else: 
            # Sub-sampling if there are too many neurons 
            if subsample_method=="random": 
                selection = subsample_random(v, selection_test,n_count_max, seed)
            elif subsample_method=="node_participation":
                selection = subsample_by_node_participation(sl, n_count_max, dim, simplex_type=simplex_type)
            elif subsample_method=="sample_simplices": 
                selection = subsample_simplices(sl,n_count_max, dim)
        return selection, selection_test

def subsample_random(v, selection_test,n_count_max, seed):
    rng = np.random.default_rng(seed)
    subsample=rng.choice(selection_test[selection_test!=v], size=n_count_max-1, replace=False)
    return np.append(v, subsample)
def subsample_by_node_participation(sl, n_count_max, dim,simplex_type="directed"):
    node, par = np.unique(sl.loc[dim], return_counts=True)
    node_par = pd.Series(par, index=node, name=dim).sort_values(ascending=False)
    selection=node_par.index[:n_count_max]
    return selection

def subsample_simplices(sl, n_count_max,dim):
    l=len(sl.loc[dim])
    selection=np.unique(sl.loc[dim][l-1:])
    i = 2
    while i < sl.loc[dim].shape[0]:
        temp=np.unique(sl.loc[dim][l-i:])
        if temp.shape[0] <= n_count_max:
            i += 1
            selection=temp
        else:
            break
    return selection
